# __init__.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access an Sqlite3 database of chess games using the apsw package.

This interface can be used if the apsw package is installed, from PyPI for
example.
"""
