# __init__.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access a Symas LMMD database of chess games using the lmdb package.

This interface can be used if the lmdb package is installed.

The lmdb package is available as a source package, from PyPI for example.

Follow the instructions to build and install lmdb.  This includes installing
Symas LMMD.
"""
