# __init__.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access Berkeley DB database of chess games using berkeleydb package.

This interface can be used if the berkeleydb package is installed.

berkeleydb is available as a source package, from PyPI for example.

Follow the instructions to build and install berkeleydb.  This includes
installing Berkeley DB.
"""
